# Copyright (c) 2018 NEC, Corp
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils


class TestCustomDriver(datasource_driver.PollingDataSourceDriver):

    def __init__(self, name='', args=None):
        super(TestCustomDriver, self).__init__(name, args)

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'test'
        result['description'] = 'This is a fake driver used for testing'
        result['config'] = datasource_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result
